# time_of_year_deltas.py quick script to take the changes in some
# stats during discrete time periods.

from arfflib_3_2 import readARFF, joinARFF, writeARFF, projectARFF
from math import log10
from statistics import mean, median, pstdev
import sys

climate = [
    'HMtempC_mean',
    'WindSpd_mean',
    'HMtempC_median',
    'WindSpd_median',
    'HMtempC_pstdv',
    'WindSpd_pstdv',
    'HMtempC_min',
    'WindSpd_min',
    'HMtempC_max',
    'WindSpd_max',
    'wndN',
    'wndNNE',
    'wndNE',
    'wndENE',
    'wndE',
    'wndESE',
    'wndSE',
    'wndSSE',
    'wndS',
    'wndSSW',
    'wndSW',
    'wndWSW',
    'wndW',
    'wndWNW',
    'wndNW',
    'wndNNW',
    'wndUNK',
]
targets = [
    'AK_All',
    'BW_All',
    'CH_All',
    'NG_All',
    'NH_All',
    'OS_All',
    'PG_All',
    'RL_All',
    'RT_All',
    'SS_All',
]
timeseries = climate + targets

keepers = ['year', 'month'] + timeseries


if __name__ == '__main__':
    if len(sys.argv) != 4:
        raise ValueError(
            'Supply input and output arff file names and timeseries key')
    inattr, indata = readARFF(sys.argv[1])
    year_col = inattr['year'][0]
    subtime = sys.argv[3]       # week or month for now 3/2023
    group_col = inattr[subtime][0]
    if 'yearday' in inattr.keys():
        keepers.append('yearday')
    if not (subtime in keepers):
        keepers.append(subtime)
    # print("DEBUG group_col for", sys.argv[3], "is", group_col)
    # indata.sort(key=lambda row : (row[group_col, row[year_col])
    attrmap_col = {}
    for t in timeseries:
        attrmap_col[t] = inattr[t][0]
    attrmap_values = {}
    for row in indata:
        for t in timeseries:
            key = (t, row[group_col], row[year_col])
            # round floats to 6 places
            if type(row[attrmap_col[t]]) == float:
                row[attrmap_col[t]] = round(row[attrmap_col[t]],6)
            attrmap_values[key] = row[attrmap_col[t]]
    # print("DEBUG attrmap_values", attrmap_values)
    sortedkeys = sorted(attrmap_values.keys()) # attrname, weekORmonth, year
    attrmap_changes = {}
    printlist_changes = []
    stats_changes = {}
    for k in sortedkeys:
        # print("DEBUG attrmap_value ",k,":",attrmap_values[k])
        attrname = k[0]
        group = k[1]
        year = k[2]
        previouskey = (attrname, group, year-1)
        change = None
        if previouskey in attrmap_values.keys():
            curval = attrmap_values[k]
            oldval = attrmap_values[previouskey]
            if (curval != None) and (oldval != None):
                change = round(curval - oldval,6)
        attrmap_changes[k] = change
        # print("DEBUG attrmap_change",k,":",attrmap_changes[k])
        # Make group and year into strings for alpha primary sort.
        printlist_changes.append([attrname + '_td', str(group), str(year),
            change])
        if len(printlist_changes[-1][1]) != 2:
            # Make month or week string have 2 digits:
            printlist_changes[-1][1] = '0' + printlist_changes[-1][1] 
        statskey = (printlist_changes[-1][0],printlist_changes[-1][1])
        if (not (statskey in stats_changes)) and (change != None):
            stats_changes[statskey] = []
        if change != None: stats_changes[statskey].append(change)
    for k in stats_changes.keys():
        stats_changes[k] = (round(mean(stats_changes[k]),6),
            round(pstdev(stats_changes[k]),6),
            round(median(stats_changes[k]),6),
            round(min(stats_changes[k]),6),
            round(max(stats_changes[k]),6))
    # Secondary sort is descending on strength (abs magnitude) of change.
    # Primary overrides!
    # printlist_changes.sort(key=lambda row :
        # (abs(row[3]) if (row[3] != None) else -1), reverse=True)
    # Primary sort is on attrname and group (week or month)
    printlist_changes.sort()
    printfile = open(sys.argv[2] + '_td.csv', 'w')
    printfile.write('attr_td,\t' + subtime
        + ',\tyear,\tchange,\tmean,\tpstdev,\tmedian,\tmin,\tmax\n')
    for row in printlist_changes:
        statskey = (row[0],row[1])
        stats = stats_changes[statskey]
        row.extend(stats)
        printfile.write(row[0])
        for index in range(1,len(row)):
            printfile.write(',\t' + str(row[index]))
        printfile.write('\n')
    printfile.close()
    # Prepare to join changes into original dataset via position joinARFF
    nameTypePairs = []
    for t in timeseries:
        nameTypePairs.append((t + '_td', 'numeric'))
        keepers.append(t + '_td')
    newdata = []
    for row in indata:
        newrow = []
        for t in timeseries:
            k = (t, row[group_col], row[year_col])
            value = attrmap_changes[k]
            newrow.append(value)
        newdata.append(newrow)
    allattrs, alldata = joinARFF(inattr, indata, nameTypePairs, newdata)
    allattrs, alldata = projectARFF(allattrs, alldata, keepers, True)
    # Tack the log10 of the targets on 3/21/2023:
    newdata = []
    nameTypePairs = []
    attrmap_col = {}
    for tgt in targets:
        attrmap_col[tgt] = allattrs[tgt][0]
        nameTypePairs.append((tgt + '_log10', 'numeric'))
    for row in alldata:
        newrow = []
        for tgt in targets:
            col = attrmap_col[tgt]
            newval = round(log10(row[col]+1),6) if (row[col] != None) else None
            newrow.append(newval)
        newdata.append(newrow)
    allattrs, alldata = joinARFF(allattrs, alldata, nameTypePairs, newdata)
    writeARFF(sys.argv[2], subtime + '_change', allattrs, alldata,
        isDebugMode=False, clobber=True)
    # Split each time band out into its own ARFF file
    subtime_col = allattrs[subtime][0]
    subtime2rows = {}
    for row in alldata:
        subtime_val = row[subtime_col]
        if not (subtime_val in subtime2rows.keys()):
            subtime2rows[subtime_val] = []
        subtime2rows[subtime_val].append(row)
    for period in sorted(subtime2rows.keys()):
        writeARFF('val' + str(period) + '_' + sys.argv[2],
            'val' + str(period) + '_' + subtime + '_change',
            allattrs, subtime2rows[period], isDebugMode=False, clobber=True)

